"""This module defines a BaseHandler class holding different models"""

from housekeeper.store import models


class BaseHandler:
    """This is a base class holding different models"""

    Bundle = models.Bundle
    Version = models.Version
    File = models.File
    Tag = models.Tag



# -*- coding: utf-8 -*-
from pathlib import Path
from typing import List

import click

from housekeeper.store import Store


@click.group()
@click.pass_context
def add(context):
    """Add things to the store."""
    context.obj["db"] = Store(context.obj["database"], context.obj["root"])


@add.command()
@click.argument("name")
@click.pass_context
def bundle(context, name):
    """Add a new bundle."""
    if context.obj["db"].bundle(name):
        click.echo(click.style("bundle name already exists", fg="yellow"))
        context.abort()
    new_bundle = context.obj["db"].new_bundle(name)
    context.obj["db"].add_commit(new_bundle)

    # add default version
    new_version = context.obj["db"].new_version(created_at=new_bundle.created_at)
    new_version.bundle = new_bundle
    context.obj["db"].add_commit(new_version)

    click.echo(
        click.style(
            f"new bundle added: {new_bundle.name} ({new_bundle.id})", fg="green"
        )
    )


@add.command("file")
@click.option("-t", "--tag", "tags", multiple=True, help="tag to associate the file by")
@click.option("-a", "--archive", is_flag=True, help="mark file to be archived")
@click.argument("bundle_name")
@click.argument("path")
@click.pass_context
def file_cmd(context, tags, archive, bundle_name, path):
    """Add a file to a bundle."""
    bundle_obj = context.obj["db"].bundle(bundle_name)
    if bundle_obj is None:
        click.echo(click.style(f"unknown bundle: {bundle_name}", fg="red"))
        context.abort()
    version_obj = bundle_obj.versions[0]
    new_file = context.obj["db"].new_file(
        path=str(Path(path).absolute()),
        to_archive=archive,
        tags=[
            context.obj["db"].tag(tag_name)
            if context.obj["db"].tag(tag_name)
            else context.obj["db"].new_tag(tag_name)
            for tag_name in tags
        ],
    )
    new_file.version = version_obj
    context.obj["db"].add_commit(new_file)
    click.echo(
        click.style(f"new file added: {new_file.path} ({new_file.id})", fg="green")
    )


@add.command()
@click.argument("tags", nargs=-1)
@click.option("-f", "--file-id", "file_id", type=int)
@click.pass_context
def tag(context: click.Context, tags: List[str], file_id: int = None):
    """Add tags to an existing file."""
    file_obj = None

    if file_id:
        file_obj = context.obj["db"].file_(file_id)
        if not file_obj:
            print(click.style("unable to find file", fg="red"))
            context.abort()

    for tag_name in tags:
        tag_obj = context.obj["db"].tag(tag_name)

        if not tag_obj:
            print(click.style(f"{tag_name}: tag created", fg="green"))
            tag_obj = context.obj["db"].new_tag(tag_name)

        if not file_obj:
            continue

        if tag_obj in file_obj.tags:
            print(click.style(f"{tag_name}: tag already added", fg="yellow"))
            continue

        file_obj.tags.append(tag_obj)

    context.obj["db"].commit()

    if not file_obj:
        context.exit()

    all_tags = (tag.name for tag in file_obj.tags)
    print(click.style(f"file tags: {', '.join(all_tags)}", fg="blue"))



# -*- coding: utf-8 -*-
"""This is a core module for the Store API"""

import logging
from pathlib import Path

import alchy

from housekeeper.store.api.add import AddHandler
from housekeeper.store.api.find import FindHandler
from housekeeper.store import models

LOG = logging.getLogger(__name__)


class CoreHandler(FindHandler, AddHandler):
    """Aggregating class for the store api handlers"""


class Store(alchy.Manager, CoreHandler):

    """
    Handles interactions with the database in the context when a temporary
    database connection is needed, e.g. a command line interface.

    Args:
        uri (str): SQLAlchemy database connection str
    """

    def __init__(self, uri: str, root: str):
        super(Store, self).__init__(config=dict(SQLALCHEMY_DATABASE_URI=uri), Model=models.Model)
        self.File.app_root = Path(root)
        self.Version.app_root = Path(root)



"""
This module handles finding things in the store/database
"""
import datetime as dt
from typing import List
from pathlib import Path
from dateutil.parser import parse as parse_date

from sqlalchemy import func as sqlalchemy_func

from housekeeper.store import models
from .base import BaseHandler


class FindHandler(BaseHandler):

    def bundles(self):
        """Fetch bundles."""
        return self.Bundle.query

    def bundle(self, name: str) -> models.Bundle:
        """Fetch a bundle from the store."""
        return self.Bundle.filter_by(name=name).first()

    def version(self, bundle: str, date: dt.datetime) -> models.Version:
        """Fetch a version from the store."""
        return (self.Version.query
                            .join(models.Version.bundle)
                            .filter(models.Bundle.name == bundle,
                                    models.Version.created_at == date)
                            .first())

    def tag(self, name: str) -> models.Tag:
        """Fetch a tag from the database."""
        return self.Tag.filter_by(name=name).first()

    def file_(self, file_id: int):
        """Get a file by record id."""
        return self.File.get(file_id)

    def files(self, *, bundle: str=None, tags: List[str]=None, version: int=None,
              path: str=None) -> models.File:
        """Fetch files from the store."""
        query = self.File.query
        if bundle:
            query = (query.join(self.File.version, self.Version.bundle)
                          .filter(self.Bundle.name == bundle))

        if tags:
            # require records to match ALL tags
            query = (
                query.join(self.File.tags)
                .filter(self.Tag.name.in_(tags))
                .group_by(models.File.id)
                .having(sqlalchemy_func.count(models.Tag.name) == len(tags))
            )

        if version:
            query = query.join(self.File.version).filter(self.Version.id == version)

        if path:
            query = query.filter_by(path=path)

        return query


    def files_before(self, *, bundle: str=None, tags: List[str]=None, before:
                     str=None) -> models.File:
        """Fetch files before date from store"""
        query = self.files(tags=tags, bundle=bundle)
        if before:
            before_dt = parse_date(before)
            query = query.join(models.Version).filter(models.Version.created_at < before_dt)

        return query


    def files_ondisk(self, file_objs: models.File) -> set:
        """Returns a list of files that are not on disk."""

        return set([file_obj for file_obj in file_objs if Path(file_obj.full_path).is_file()])



# -*- coding: utf-8 -*-
"""This module handles adding things to the store"""

import logging
from pathlib import Path
from typing import List
import datetime as dt

from housekeeper.store import models
from housekeeper.store.api.base import BaseHandler
from housekeeper.store.api.find import FindHandler


LOG = logging.getLogger(__name__)


class AddHandler(BaseHandler):
    """Handles adding things to the store"""

    def __init__(self):
        super().__init__()
        AddHandler.version = FindHandler.version
        AddHandler.bundle = FindHandler.bundle
        AddHandler.tag = FindHandler.tag

    def add_bundle(self, data: dict) -> (models.Bundle, models.Version):
        """Build a new bundle version of files.

        The format of the input dict is defined in the `schema` module.
        """
        bundle_obj = self.bundle(data["name"])
        if bundle_obj and self.version(bundle_obj.name, data["created"]):
            LOG.debug("version of bundle already added")
            return None

        if bundle_obj is None:
            bundle_obj = self.new_bundle(name=data["name"], created_at=data["created"])

        version_obj = self.new_version(
            created_at=data["created"], expires_at=data.get("expires")
        )

        tag_names = set(
            tag_name for file_data in data["files"] for tag_name in file_data["tags"]
        )
        tag_map = self._build_tags(tag_names)

        for file_data in data["files"]:
            if isinstance(file_data["path"], str):
                paths = [file_data["path"]]
            else:
                paths = file_data["path"]
            for path in paths:
                LOG.debug("adding file: %s", path)
                if not Path(path).exists():
                    raise FileNotFoundError(path)
                tags = [tag_map[tag_name] for tag_name in file_data["tags"]]
                new_file = self.new_file(
                    path, to_archive=file_data["archive"], tags=tags
                )
                version_obj.files.append(new_file)

        version_obj.bundle = bundle_obj
        return bundle_obj, version_obj

    def _build_tags(self, tag_names: List[str]) -> dict:
        """Build a list of tag objects."""
        tags = {}
        for tag_name in tag_names:
            tag_obj = self.tag(tag_name)
            if tag_obj is None:
                LOG.debug("create new tag: %s", tag_name)
                tag_obj = self.new_tag(tag_name)
            tags[tag_name] = tag_obj
        return tags

    def new_bundle(self, name: str, created_at: dt.datetime = None) -> models.Bundle:
        """Create a new file bundle."""
        new_bundle = self.Bundle(name=name, created_at=created_at)
        return new_bundle

    def new_version(
        self, created_at: dt.datetime, expires_at: dt.datetime = None
    ) -> models.Version:
        """Create a new bundle version."""
        new_version = self.Version(created_at=created_at, expires_at=expires_at)
        return new_version

    def new_file(
        self,
        path: str,
        checksum: str = None,
        to_archive: bool = False,
        tags: List[models.Tag] = None,
    ) -> models.File:
        """Create a new file."""
        new_file = self.File(
            path=path, checksum=checksum, to_archive=to_archive, tags=tags
        )
        return new_file

    def new_tag(self, name: str, category: str = None) -> models.Tag:
        """Create a new tag."""
        new_tag = self.Tag(name=name, category=category)
        return new_tag
